#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""A logging handler file, which helps in providing logs of the script execution."""
from logging.handlers import RotatingFileHandler
from datetime import datetime
from platform import system
import logging
import os

WORK_PATH = os.path.abspath(os.getcwd())
now = datetime.now()
LOGGER = ""

logger = logging.getLogger(__name__)
formatting = logging.Formatter("%(asctime)s %(levelname)s: %(message)s [in %(pathname)s:%(lineno)d]")

if system() == "Linux" or system() == "Darwin":
    LOGGER += WORK_PATH + "{}".format("/logs")
    if not os.path.exists(LOGGER):
        os.mkdir(LOGGER)
    handler = RotatingFileHandler("{}/{}".format(LOGGER, "app.log"), maxBytes=1000000, backupCount=20)
    handler.setFormatter(formatting)
    logger.addHandler(handler)

if system() == "Windows":
    LOGGER += WORK_PATH + "{}".format("\\logs")
    if not os.path.exists(LOGGER):
        os.mkdir(LOGGER)
    handler = RotatingFileHandler("{}\\{}".format(LOGGER, "app.log"), maxBytes=1000000, backupCount=20)
    handler.setFormatter(formatting)
    logger.addHandler(handler)


def add_log(message, level):
    """Writes a log to a log file with activity done."""
    if level == "debug".lower():
        logger.setLevel(logging.DEBUG)
        logger.debug(message)
    elif level == "error".lower():
        logger.setLevel(logging.ERROR)
        logger.error(message)
    else:
        logger.setLevel(logging.INFO)
        logger.info(message)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Contains all Classes used to Authenticate Jira, endpoints and PrettyPrint.

- Instantiate the endpoint to Jira API
- alias to Authentication
- alias to Endpoints
- friendly name to PrettyPrint
"""
from requests.auth import HTTPBasicAuth
from typing import Any, Optional
from pprint import PrettyPrinter
import requests


class Credentials(object):
    """class.Credentials -> used for authentication of the user to the Instance."""

    auth_request = None
    headers = None

    def __init__(self, user: str, password: str, url: str = Any) -> Optional:
        """Instantiate the login."""
        self.base_url = url
        self.token_session(email=user, token=password)

    # produce a session for the script and save the session
    def token_session(self, email: Any = None, token: Any = None) -> Any:
        self.auth_request = HTTPBasicAuth(email, token)
        self.headers = {"Content-Type": "application/json"}

    def get(self, url, *args, payload=None, **kwargs):
        response = requests.get(url, *args, auth=self.auth_request, json=payload, headers=self.headers, **kwargs)
        return response

    def post(self, url, *args, payload=None, **kwargs):
        response = requests.post(url, *args, auth=self.auth_request, json=payload, headers=self.headers, **kwargs)
        return response

    def put(self, url, *args, payload=None, **kwargs):
        response = requests.put(url, *args, auth=self.auth_request, json=payload, headers=self.headers, **kwargs)
        return response

    def delete(self, url, **kwargs):
        response = requests.delete(url, auth=self.auth_request, headers=self.headers, **kwargs)
        return response


class Echo(PrettyPrinter):
    """A Class used to inherit from PrettyPrinter."""

    def __init__(self, *args, **kwargs):
        """Inherit from the parent."""
        super(Echo, self).__init__(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        return self.__init__(*args, **kwargs)

    def echo(self, raw):
        """Prints the formatted representation of object on stream, followed by a newline."""
        return self.pprint(object=raw)


class InitProcess(Credentials):
    """should inherit an instance of Credential class using super().

    Object values are entered directly when called because of the __call__
    dunder method."""

    def __init__(self, user=None, password=None, url=None):
        """A Call to the Credential Class."""
        super(InitProcess, self).__init__(user=user, password=password, url=url)

    def __call__(self, *args, **kwargs):
        return self.__init__(*args, **kwargs)


LOGIN = InitProcess()


class EndPoints:
    """A Structural way to dynamically load urls that is fed to other functions."""

    @classmethod
    def myself(cls):
        """Return data on your own user."""
        return "{}/rest/api/3/myself".format(LOGIN.base_url)

    @classmethod
    def search_users(cls, query: int = 0):
        """Search multiple users and retrieve the data

        :param: startAt
        """
        return "{}/rest/api/3/users/search?startAt={}&maxResults=50".format(LOGIN.base_url, query)

    @classmethod
    def get_user_group(cls, account_id: Any):
        """Search for the groups a user belongs to

        :param: accountId required
        """
        return f"{LOGIN.base_url}/rest/api/3/user/groups?accountId={account_id}"

    @classmethod
    def get_projects(cls, *args: Any, start_at=0, max_results=50):
        """Return a list of Projects available on an Instance

        How to use this endpoint /rest/api/3/project/search  is mentioned here
        1. https://developer.atlassian.com/cloud/jira/platform/rest/v3/api-group-projects/
           #api-rest-api-3-project-search-get
        2. Query Parameters that are useful mostly.
           a) query, example: query=key,name {caseInsensitive}
           b) searchBy, example: searchBy=key,name
           c) action, example: action=browse
               i. available options [view, browse, edit]
           d) status example: status=live
               i. available options [live, archived, deleted]
           e). expand, example: expand=insight
                 i. available options [insight, description, projectKeys, url, issueTypes, lead]
        3. startAt defaults as keyword args, example: startAt=0
        4. maxResults defaults as keyword args, example: maxResults=50
        """
        param = []
        amp = "&"
        num_list = 0
        if args:
            nos = len(args)
            if nos > 0:
                for v in args:
                    param.append(f"{v}{amp}")
                    num_list += 1
                    if num_list >= nos:
                        # cut the last ampersand and clear the list
                        cut = "".join(param).rstrip("&")
                        param.clear()
                        param.append(cut)
                        break
                print("Project Search Query Parameter:", param[0])
                return "{}/rest/api/3/project/search?{}&startAt={}&maxResults={}"\
                    .format(LOGIN.base_url, param[0], start_at, max_results)
        else:
            return "{}/rest/api/3/project/search?startAt={}&maxResults={}"\
                .format(LOGIN.base_url, start_at, max_results)

    @classmethod
    def find_users_with_permission(cls, *args):
        """Find users with permissions to a Project

        :param: 1st accountId
        :param: 2nd projectKey
        :param: 3rd permissions that needs checking all in caps e.g "BROWSE", "CREATE_ISSUE" etc
        """
        return "{}/rest/api/3/user/permission/search?accountId={}&projectKey={}&permissions={}" \
            .format(LOGIN.base_url, *args)

    @classmethod
    def get_roles_for_project(cls, id_or_key: Any):
        """Returns a list of project roles for the project returning the name and self URL for each role.

        :param: projectKey or Id
        """
        return "{}/rest/api/3/project/{}/role".format(LOGIN.base_url, id_or_key)

    @classmethod
    def get_project_role(cls, *args):
        """Returns a project role's details and actors associated with the project.

        :param: projectKey or Id of the Project
        :param: id of the role
        """
        return "{}/rest/api/3/project/{}/role/{}".format(LOGIN.base_url, *args)

    @classmethod
    def get_all_permission_scheme(cls, query: str = None):
        """Returns all permission schemes."""
        if query is not None:
            return "{}/rest/api/3/permissionscheme?{}".format(LOGIN.base_url, query)
        else:
            return "{}/rest/api/3/permissionscheme".format(LOGIN.base_url)

    @classmethod
    def get_all_issue_type_schemes(cls, query: Optional[str] = None, start_at=0, max_results=50):
        """Returns a paginated list of issue type schemes.

        Only issue type schemes used in classic projects are returned"""
        if query is not None:
            return "{}/rest/api/3/issuetypescheme?{}&startAt={}&maxResults={}".format(LOGIN.base_url, query,
                                                                                      start_at, max_results)
        else:
            return "{}/rest/api/3/issuetypescheme?startAt={}&maxResults={}".format(LOGIN.base_url,
                                                                                   start_at, max_results)

    @classmethod
    def get_all_issue_types(cls):
        """Returns all issue types.

        if the user has the Administer Jira global permission, all issue types are returned.
        if the user has the Browse projects project permission for one or more projects,
        the issue types associated with the projects the user has permission to browse are returned.
        """
        return "{}/rest/api/3/issuetype".format(LOGIN.base_url)

    @classmethod
    def get_all_issue_security_scheme(cls):
        """Returns all issue security schemes."""
        return "{}/rest/api/3/issuesecurityschemes".format(LOGIN.base_url)

    @classmethod
    def get_all_priorities(cls):
        """Returns the list of all issue priorities."""
        return "{}/rest/api/3/priority".format(LOGIN.base_url)

    @classmethod
    def search_all_notification_schemes(cls, query: Optional[str] = None, start_at=0, max_results=50):
        """Returns a paginated list of notification schemes ordered by display name.

        :param: 1st String value for expand=: {all, field, group, user, projectRole, notificationSchemeEvents}
        :param: 2nd startAt=0 has default value
        :param: 3rd maxResults=50 has default value
        """
        if query is not None:
            return "{}/rest/api/3/notificationscheme?{}&startAt={}&maxResults={}".format(LOGIN.base_url, query,
                                                                                         start_at, max_results)
        else:
            return "{}/rest/api/3/notificationscheme?startAt={}&maxResults={}".format(LOGIN.base_url,
                                                                                      start_at, max_results)

    @classmethod
    def get_field(cls, query: Optional[str] = None, start_at: int = 0, max_results: int = 50):
        """Returns a paginated list of fields for Classic Jira projects. The list can include:

        *  all fields.
        *  specific fields, by defining id.
        *  fields that contain a string in the field name or description, by defining query.
        *  specific fields that contain a string in the field name or description, by defining id and query.
        Only custom fields can be queried, type must be set to custom.
         """
        if query is not None:
            return "{}/rest/api/3/field/search?{}&startAt={}&maxResults={}".format(LOGIN.base_url, query,
                                                                                   start_at, max_results)
        else:
            return "{}/rest/api/3/field/search?startAt={}&maxResults={}".format(LOGIN.base_url,
                                                                                start_at, max_results)

    @classmethod
    def get_attachment_meta_data(cls, query: str, warning: Any = None):
        """Returns the metadata for an attachment. Note that the attachment itself is not returned.

         :param: id of the attachment
         Use issue search endpoint in conjunction to grab the attachment id
         """
        import warnings
        message = "We will be removing this endpoint soon\n use endpoint.issue_attachments() instead."
        if warning is None:
            warnings.warn(message, DeprecationWarning, stacklevel=2)
        else:
            ...
        return "{}/rest/api/3/attachment/{}".format(LOGIN.base_url, query)

    @classmethod
    def issue_attachments(cls, id_or_key: str = None, attach_id: str = None, uri: Optional[str] = None,
                          query: Optional[str] = None):
        """Returns the attachment content.

        :request: GET - Get Jira attachment settings
        Returns the attachment settings, that is, whether attachments are enabled and the
        maximum attachment size allowed.

                 GET - Get attachment Meta data
        Returns the metadata for an attachment. Note that the attachment itself is not returned.

        :param: attach_id required (id of the attachment), datatype -> string
                DELETE - Deletes an attachment from an issue.

        param: attach_id required (id of the attachment), datatype -> string
               GET - Get all metadata for an expanded attachment
               :param: query, datatype -> string
               available options
                * expand/human -Returns the metadata for the contents of an attachment, if it is an archive,
                     and metadata for the attachment itself. For example, if the attachment is a ZIP archive,
                     then information about the files in the archive is returned and metadata for the ZIP archive.

                * expand/raw - Returns the metadata for the contents of an attachment, if it is an archive.
                     For example, if the attachment is a ZIP archive, then information about the files in the
                     archive is returned. Currently, only the ZIP archive format is supported.

               POST - Adds one or more attachments to an issue. Attachments are posted as multipart/form-data
        POST - Adds one or more attachments to an issue. Attachments are posted as multipart/form-data
        :param: id_or_key required, datatype -> string
        The ID or key of the issue that attachments are added to.
        """
        if uri is not None:
            return "{}/rest/api/3/attachment/{}".format(LOGIN.base_url, uri)
        else:
            if query is not None and attach_id is not None and id_or_key is None:
                return "{}/rest/api/3/attachment/{}/{}".format(LOGIN.base_url, attach_id, query)
            elif query is not None and attach_id is None and id_or_key is not None:
                return "{}/rest/api/3/attachment/{}/{}".format(LOGIN.base_url, id_or_key, query)
            else:
                return "{}/rest/api/3/attachment/{}".format(LOGIN.base_url, attach_id)

    @classmethod
    def search_issues_jql(cls, query, start_at: int = 0, max_results: int = 50):
        """Searches for issues using JQL."""
        return "{}/rest/api/3/search?jql={}&startAt={}&maxResults={}".format(LOGIN.base_url, query,
                                                                             start_at, max_results)

    @classmethod
    def search_for_filters(cls, query: Optional[str] = None, start_at: int = 0):
        """Returns a paginated list of filters. Use this operation to get:

        *  specific filters, by defining id only.
        *  filters that match all of the specified attributes. For example, all filters
           for a user with a particular word in their name. When multiple attributes are
           specified only filters matching all attributes are returned.
        :param: 1st String value: filterName, accountId, owner, groupname, projectId
        :param: 2nd startAt=0 has default value
        :param: filled - maxResults=50 (default)
         """
        if query is not None:
            return "{}/rest/api/3/filter/search?{}&startAt={}&maxResults=50".format(LOGIN.base_url, query, start_at)
        else:
            return "{}/rest/api/3/filter/search?startAt={}&maxResults=50".format(LOGIN.base_url, start_at)

    @classmethod
    def search_for_dashboard(cls, query: Optional[str] = None, start_at: int = 0):
        """Returns a paginated list of dashboards. This operation is similar to

        Get dashboards except that the results can be refined to include dashboards that
        have specific attributes. For example, dashboards with a particular name.
        When multiple attributes are specified only filters matching all attributes
        are returned.
        :param: 1st String value: dashboardName, accountId, owner, groupname, projectId
        :param: 2nd startAt=0 has default value
        :param: filled - maxResult=20 (default)
        """
        if query is not None:
            return "{}/rest/api/3/dashboard/search?{}&startAt={}&maxResults=20".format(LOGIN.base_url, query, start_at)
        else:
            return "{}/rest/api/3/dashboard/search?startAt={}&maxResults=20".format(LOGIN.base_url, start_at)

    @classmethod
    def get_dashboard(cls, dashboard_id: int):
        """Gets the dashboard."""
        return "{}/rest/api/3/dashboard/{}".format(LOGIN.base_url, dashboard_id)

    @classmethod
    def get_all_application_role(cls):
        """
        Returns all application roles.

        In Jira, application roles are managed using the Application access configuration page."""
        return "{}/rest/api/3/applicationrole".format(LOGIN.base_url)

    @classmethod
    def search_all_workflows(cls, query: int = 0):
        """
        Returns a paginated list of published classic workflows. When workflow names are specified.

        details of those workflows are returned. Otherwise, all published classic workflows are returned.
        This operation does not return next-gen workflows.
        :param: 1st startAt=0 has default value
        :param: filled - maxResults=50 (default)
        """
        return "{}/rest/api/3/workflow/search?startAt={}&maxResults=50".format(LOGIN.base_url, query)

    @classmethod
    def search_all_workflow_schemes(cls, query: int = 0):
        """Returns a paginated list of all workflow schemes, not including draft workflow schemes.

        :param: 1st startAt=0 has default value
        :param: filled - maxResults=50 (default)
        """
        return "{}/rest/api/3/workflowscheme?startAt={}&maxResults=50".format(LOGIN.base_url, query)

    @classmethod
    def search_all_screens(cls, query: int = 0):
        """Returns a paginated list of all screens or those specified by one or more screen IDs.

        :param: 1st startAt=0 has default value
        :param: maxResults=100 (default)
        """
        return "{}/rest/api/3/screens?startAt={}&maxResults=100".format(LOGIN.base_url, query)

    @classmethod
    def search_for_screen_schemes(cls, query: int = 0):
        """Returns a paginated list of screen schemes.

            Only screen schemes used in classic projects are returned.
        :param: 1st - startAt=0 has default value
        :param: maxResults=25 (default)
        """
        return "{}/rest/api/3/screenscheme?startAt={}&maxResults=25".format(LOGIN.base_url, query)

    @classmethod
    def get_project_component(cls, id_or_key):
        """Returns all components in a project. See the Get project components paginated

            resource if you want to get a full list of components with pagination.
            The project ID or project key (case sensitive).
        """
        return "{}/rest/api/3/project/{}/components".format(LOGIN.base_url, id_or_key)

    @classmethod
    def get_resolutions(cls):
        """Returns a list of all issue resolution values."""
        return "{}/rest/api/3/resolution".format(LOGIN.base_url)

    ################################################
    # Jira Software Specifics API endpoints
    ################################################
    # BACKLOG -> API for backlog
    @classmethod
    def move_issues_to_backlog(cls):
        """Move issues to the backlog.

        This operation is equivalent to remove future and active sprints from a given set of issues.
        At most 50 issues may be moved at once.

        :request:POST:
        :body param: issues, datatype -> Array<string>
        Send a POST request within API.
        """
        return "{}/rest/agile/1.0/backlog/issue".format(LOGIN.base_url)

    @classmethod
    def move_issues_to_backlog_from_board(cls, board_id):
        """Move issues to the backlog of a particular board (if they are already on that board).

        This operation is equivalent to remove future and active sprints from a given set of issues
        if the board has sprints If the board does not have sprints this will put the issues back
        into the backlog from the board. At most 50 issues may be moved at once.

        :request:POST:
        :param: board_id required
        :body param: issues, datatype -> Array<string>,
                   : rankBeforeIssue, rankAfterIssue, type -> string
                   : rankCustomFieldId, type -> integer
        """
        return "{}/rest/agile/1.0/backlog/{}/issue".format(LOGIN.base_url, board_id)

    # BOARD -> API for Boards

    @classmethod
    def create_board(cls):
        """Creates a new board. Board name, type and filter ID is required.

        :request:POST:
        :body param: name, type, datatype -> string
                   : filterId, datatype -> integer
                   : location, datatype -> object
        """
        return "{}/rest/agile/1.0/board".format(LOGIN.base_url)

    @classmethod
    def get_board_by_filter_id(cls, filter_id, start_at: int = 0, max_results: int = 50) -> Any:
        """Returns any boards which use the provided filter id.

        This method can be executed by users without a valid software license
        in order to find which boards are using a particular filter.

        :param: filter_id  required
        Filters results to boards that are relevant to a filter.
        Not supported for next-gen boards.

        :query param: start_at defaults to 0
                    : max_results defaults to 50
        """
        return "{}/rest/agile/1.0/board/filter/{}?startAt={}&maxResults={}"\
            .format(LOGIN.base_url, filter_id, start_at, max_results)

    @classmethod
    def get_board(cls, board_id):
        """Returns the board for the given board ID.

        This board will only be returned if the user has permission to view it.
        Admins without the view permission will see the board as a private one,
        so will see only a subset of the board's data (board location for instance).
        """
        return "{}/rest/agile/1.0/board/{}".format(LOGIN.base_url, board_id)

    @classmethod
    def get_issues_on_backlog(cls, board_id, query: str = None, start_at: int = 0, max_results: int = 50) -> Any:
        """Returns all issues from the board's backlog, for the given board ID.

        This only includes issues that the user has permission to view.
        The backlog contains incomplete issues that are not assigned to any future or active sprint.

        :param: board_id required
        :query param: start_at defaults to 0,
                    : max_results defaults to 50
                    : query -> includes other query parameters such as
                         Query           Datatypes
                        ----------------------------
                         jql           | string
                         validateQuery | boolean
                         fields        | Array<string>
                         expand        | string
        """
        if query is not None:
            return "{}/rest/agile/1.0/board/{}/backlog?{}&startAt={}&maxResults={}"\
                .format(LOGIN.base_url, board_id, query, start_at, max_results)
        else:
            return "{}/rest/agile/1.0/board/{}/backlog?startAt={}&maxResults={}"\
                .format(LOGIN.base_url, board_id, start_at, max_results)

    @classmethod
    def get_issues_on_board(cls, board_id, query: str = None, start_at: int = 0, max_results: int = 50) -> Any:
        """Returns all issues from a board, for a given board ID.

        This only includes issues that the user has permission to view.
        An issue belongs to the board if its status is mapped to the board's column.

        :param: board_id required
        :query param: start_at defaults to 0,
                    : max_results defaults to 50
                    : query -> includes other query parameters such as
                         Query           Datatypes
                        ----------------------------
                         jql           | string
                         validateQuery | boolean
                         fields        | Array<string>
                         expand        | string
        """
        if query is not None:
            return "{}/rest/agile/1.0/board/{}/issue?{}&startAt={}&maxResults={}"\
                .format(LOGIN.base_url, board_id, query, start_at, max_results)
        else:
            return "{}/rest/agile/1.0/board/{}/issue?startAt={}&maxResults={}"\
                .format(LOGIN.base_url, board_id, start_at, max_results)

    @classmethod
    def move_issues_to_board(cls, board_id):
        """Move issues from the backog to the board (if they are already in the backlog of that board).

        This operation either moves an issue(s) onto a board from the backlog (by adding it to the issueList
        for the board) Or transitions the issue(s) to the first column for a kanban board with backlog.

        :request:POST:
        :param: board_id required
        :body param: rankBeforeIssue, rankAfterIssue, datatype -> string
                   : rankCustomFieldId, datatype -> integer
                   : issues, datatype -> Array<string>
        """
        return "{}/rest/agile/1.0/board/{}/issue".format(LOGIN.base_url, board_id)

    @classmethod
    def get_projects_on_board(cls, board_id, start_at: int = 0, max_results: int = 50) -> Any:
        """Returns all projects that are associated with the board, for the given board ID.

         If the user does not have permission to view the board, no projects will be returned at all.
         Returned projects are ordered by the name.

         :param: board_id required
         :query param: start_at defaults 0
                     : max_results defaults 50
        """
        return "{}/rest/agile/1.0/board/{}/project?startAt={}&maxResults={}"\
            .format(LOGIN.base_url, board_id, start_at, max_results)

    @classmethod
    def get_all_quick_filters(cls, board_id, start_at: int = 0, max_results: int = 50) -> Any:
        """Returns all quick filters from a board, for a given board ID.

        :param: board_id required
        :query param: start_at defaults 0
                     : max_results defaults 50
        """
        return "{}/rest/agile/1.0/board/{}/quickfilter?startAt={}&maxResults={}"\
            .format(LOGIN.base_url, board_id, start_at, max_results)

    @classmethod
    def get_quick_filter(cls, board_id, quick_filter_id):
        """Returns the quick filter for a given quick filter ID.

        The quick filter will only be returned if the user can view the board that the
        quick filter belongs to.

        :param: board_id required,
                quick_filter_id required
        """
        return "{}/rest/agile/1.0/board/{}/quickfilter/{}".format(LOGIN.base_url, board_id, quick_filter_id)

    @classmethod
    def get_all_sprints(cls, board_id, query: str = None, start_at: int = 0, max_results: int = 50) -> Any:
        """Get all Sprint on a Board."""
        if query is not None:
            return "{}/rest/agile/1.0/board/{}/sprint?startAt={}&maxResults={}"\
                .format(LOGIN.base_url, board_id, query, start_at, max_results)
        else:
            return "{}/rest/agile/1.0/board/{}/sprint?startAt={}&maxResults={}"\
                .format(LOGIN.base_url, board_id, start_at, max_results)

    # SPRINT -> API for Sprints
    @classmethod
    def create_sprint(cls):
        """Creates a future sprint. Sprint name and origin board id are required.

        Start date, end date, and goal are optional.
        :request:POST:
        :body param: name, startDate, endDate, goal, datatype -> string
                   : originBoardId, datatype -> integer
        """
        return "{}/rest/agile/1.0/sprint".format(LOGIN.base_url)

    @classmethod
    def get_sprint(cls, sprint_id):
        """Returns the sprint for a given sprint ID.

        The sprint will only be returned if the user can view the board that the sprint was created on,
        or view at least one of the issues in the sprint.
        :param: sprint_id required
        """
        return "{}/rest/agile/1.0/sprint/{}".format(LOGIN.base_url, sprint_id)

    @classmethod
    def update_sprint(cls, sprint_id):
        """Performs a full update of a sprint.

        A full update means that the result will be exactly the same as the request body.
        Any fields not present in the request JSON will be set to null.

        :request:PUT:
        :param: sprint_id required
        :body param: name, state, startDate, endDate, goal, self (format: uri), completeDate, datatype -> string
                   : id, originBoardId, datatype -> integer
        """
        return "{}/rest/agile/1.0/sprint/{}".format(LOGIN.base_url, sprint_id)

    @classmethod
    def delete_sprint(cls, sprint_id):
        """Deletes a sprint.

        Once a sprint is deleted, all open issues in the sprint will be moved to the backlog.
        :request:DELETE:
        :param: sprint_id required
        """
        return "{}/rest/agile/1.0/sprint/{}".format(LOGIN.base_url, sprint_id)

    ################################################
    # Jira Service Management Specific API endpoints
    ################################################
    @classmethod
    def create_customer(cls):
        """This method adds a customer to the Jira Service Management.

        instance by passing a JSON file including an email address and display name.
        :request:POST
        :body param: email, displayName, datatype -> string
        """
        return "{}/rest/servicedeskapi/customer".format(LOGIN.base_url)

    @classmethod
    def get_server_info(cls):
        """This method retrieves information about the Jira Service Management.

        instance such as software version, builds, and related links."""
        return "{}/rest/servicedeskapi/info".format(LOGIN.base_url)

    @classmethod
    def get_organizations(cls, start: int = 0, limit: int = 50, account_id: str = None) -> Any:
        """This method returns a list of organizations in the Jira Service Management instance.

        Use this method when you want to present a list of organizations or want to locate an organization by name.
        :query param: start defaults to 0
                    : limit defaults to 50
                    : account_id, datatype string. e.g. 5b10ac8d82e05b22cc7d4ef5
        """
        if account_id is not None:
            return "{}/rest/servicedeskapi/organization?{}&start={}&limit={}"\
                .format(LOGIN.base_url, account_id, start, limit)
        else:
            return "{}/rest/servicedeskapi/organization?start={}&limit={}".format(LOGIN.base_url, start, limit)

    @classmethod
    def create_organization(cls):
        """This method creates an organization by passing the name of the organization.

        :request:POST:
        :body param: name, datatype -> string
        """
        return "{}/rest/servicedeskapi/organization".format(LOGIN.base_url)

    @classmethod
    def get_organization(cls, org_id):
        """This method returns details of an organization.

        Use this method to get organization details whenever your application component is passed an organization ID
        but needs to display other organization details.
        :param: org_id required
        """
        return "{}/rest/servicedeskapi/organization/{}".format(LOGIN.base_url, org_id)

    @classmethod
    def delete_organization(cls, org_id):
        """This method deletes an organization.

        Note that the organization is deleted regardless of other associations it may have.
        For example, associations with service desks.
        :request:DELETE:
        :param: org_id required
        """
        return "{}/rest/servicedeskapi/organization/{}".format(LOGIN.base_url, org_id)

    @classmethod
    def get_users_in_organization(cls, org_id, start: int = 0, limit: int = 50) -> Any:
        """This method returns all the users associated with an organization.

        Use this method where you want to provide a list of users for an organization
        or determine if a user is associated with an organization.
        :param: org_id required
        """
        return "{}/rest/servicedeskapi/organization/{}/user?start={}&limit={}".format(LOGIN.base_url, org_id,
                                                                                      start, limit)

    @classmethod
    def add_users_to_organization(cls, org_id) -> Any:
        """This method adds users to an organization.

        :request:POST:
        :param: org_id required
        :body param: usernames, accountIds, datatypes -> Array<string>
        """
        return "{}/rest/servicedeskapi/organization/{}/user".format(LOGIN.base_url, org_id)

    @classmethod
    def remove_users_from_organization(cls, org_id) -> Any:
        """This method removes users from an organization.

        :request:DELETE:
        :param: org_id required
        :body param: usernames, accountIds, datatypes -> Array<string>
        """
        return "{}/rest/servicedeskapi/organization/{}/user".format(LOGIN.base_url, org_id)

    @classmethod
    def get_sd_organizations(cls, service_desk_id, start: int = 0, limit: int = 50, account_id: str = None):
        """This method returns a list of all organizations associated with a service desk.

        :param: service_desk_id required
        :query param: start defaults to 0
                    : limit defaults to 50
                    : account_id, datatype string. e.g. 5b10ac8d82e05b22cc7d4ef5
        """
        if account_id is not None:
            return "{}/rest/servicedeskapi/servicedesk/{}/organization?{}&start={}&limit={}"\
                .format(LOGIN.base_url, account_id, service_desk_id, start, limit)
        else:
            return "{}/rest/servicedeskapi/servicedesk/{}/organization?start={}&limit={}"\
                .format(LOGIN.base_url, service_desk_id, start, limit)

    @classmethod
    def add_sd_organization(cls, service_desk_id):
        """This method adds an organization to a service desk.

        If the organization ID is already associated with the service desk,
        no change is made and the resource returns a 204 success code.
        :request:POST:
        :param: service_desk_id required
        :body param: organizationId, datatype -> integer
        """
        return "{}/rest/servicedeskapi/servicedesk/{}/organization".format(LOGIN.base_url, service_desk_id)

    @classmethod
    def remove_sd_organization(cls, service_desk_id):
        """This method removes an organization from a service desk.

        If the organization ID does not match an organization associated with the service desk,
        no change is made and the resource returns a 204 success code.
        :request:DELETE:
        :param: service_desk_id required
        :body param: organizationId, datatype -> integer
        """
        return "{}/rest/servicedeskapi/servicedesk/{}/organization".format(LOGIN.base_url, service_desk_id)

    # SERVICEDESK -> API specific to servicedesk
    @classmethod
    def get_customers(cls, service_desk_id, start: int = 0, limit: int = 50, query: str = None):
        """This method returns a list of the customers on a service desk.

        The returned list of customers can be filtered using the query parameter.
        The parameter is matched against customers' displayName, name, or email.
        This API is experimental
        :param: service_desk_id required
        :query param: start defaults to 0
                    : limit defaults to 50
                    : query, datatype string.
        """
        if query is not None:
            return "{}/rest/servicedeskapi/servicedesk/{}/customer?{}&start={}&limit={}"\
                .format(LOGIN.base_url, query, service_desk_id, start, limit)
        else:
            return "{}/rest/servicedeskapi/servicedesk/{}/customer?start={}&limit={}"\
                .format(LOGIN.base_url, service_desk_id, start, limit)

    @classmethod
    def add_customers(cls, service_desk_id):
        """Adds one or more customers to a service desk.

        If any of the passed customers are associated with the service desk,
        no changes will be made for those customers and the resource returns a 204 success code.
        :request:POST:
        :param: service_desk_id required
        :body param: usernames, accountIds,  datatype -> Array<string>
                """
        return "{}/rest/servicedeskapi/servicedesk/{}/customer".format(LOGIN.base_url, service_desk_id)

    @classmethod
    def remove_customers(cls, service_desk_id):
        """This method removes one or more customers from a service desk.

        The service desk must have closed access. If any of the passed customers are not associated with
        the service desk, no changes will be made for those customers and the resource returns a 204 success code.
        :request:DELETE:
        :param: service_desk_id required
        :body param: usernames, accountIds,  datatype -> Array<string>
                """
        return "{}/rest/servicedeskapi/servicedesk/{}/customer".format(LOGIN.base_url, service_desk_id)

    ################################################
    # Jira Specific API endpoints
    ################################################
    @classmethod
    def jira_user(cls, account_id: str = None):
        """API for User creation, deletion and retrieval.

        :request:POST - Creates a user. This resource is retained for legacy compatibility.
                        As soon as a more suitable alternative is available this resource will be deprecated
                        :body param: key, name, password, emailAddress, displayName, notification, datatypes -> string
                                    : applicationKeys, datatype -> Array<string>
                                    : Additional Properties, datatypes -> Any
                        returns 201 for successful creation
                :DELETE - Deletes a user.
                          :body param: accountId, datatype -> string required
                          returns 204 for successful deletion
                :GET - Returns a user.
                       :body param: accountId, expand, datatypes -> string
        """
        if account_id is not None:
            return "{}/rest/api/3/user?accountId={}".format(LOGIN.base_url, account_id)
        else:
            return "{}/rest/api/3/user".format(LOGIN.base_url)

    @classmethod
    def jira_group(cls, group_name: str = None, swap_group: str = None):
        """Used for Creation and deletion of Jira groups.

        :request: POST - Creates a group.
                     :body param: name required, datatype -> string
                     returns 201 if successful
                : DELETE - Deletes a group.
                     The group to transfer restrictions to. Only comments and worklogs are transferred.
                     If restrictions are not transferred, comments and worklogs are inaccessible after the deletion.
                     :query param: group_name required, swap_group,  datatype -> string
                     returns 200 if successful
        """
        if group_name is not None and swap_group is None:
            return "{}/rest/api/3/group?groupname={}".format(LOGIN.base_url, group_name)
        elif group_name is not None and swap_group is not None:
            return "{}/rest/api/3/group?groupname={}&swapGroup={}".format(LOGIN.base_url, group_name, swap_group)
        else:
            return "{}/rest/api/3/group".format(LOGIN.base_url)

    @classmethod
    def group_jira_users(cls, group_name: str, account_id: str = None):
        """Used for addition and removal of users to and from groups.

        :request: POST - Adds a user to a group.
                     :query param: groupname required, datatype -> string
                     :body param: name, accountId, datatype -> string
                     returns 201 if successful
                : DELETE - Removes a user from a group.
                     :query param: group_name required, account_id required,  datatype -> string
                     returns 200 if successful
        """
        if account_id is not None:
            return "{}/rest/api/3/group/user?groupname={}&accountId={}".format(LOGIN.base_url, group_name, account_id)
        else:
            return "{}/rest/api/3/group/user?groupname={}".format(LOGIN.base_url, group_name)

    @classmethod
    def projects(cls, id_or_key, query: Optional[str] = None, uri: Optional[str] = None,
                 enable_undo: Optional[bool] = None) -> str:
        """Create, delete, update, archive, get status.

        :request: POST - for project creations.
                         The project types are available according to the installed Jira features as follows:
        :param: id_or_key required
              : uri optional for accessing other project endpoints -> string
                   endpoint: /rest/api/3/project/{projectIdOrKey}/{archive}
                   available options [archive, delete, restore, statuses]
                         archive - Archives a project. Archived projects cannot be deleted.
                         delete - Deletes a project asynchronously.
                         restore - Restores a project from the Jira recycle bin.
                         statuses - Returns the valid statuses for a project.

        see:https://developer.atlassian.com/cloud/jira/platform/rest/v3/api-group-projects/#api-rest-api-3-project-post

                  :body param: projectTypeKey and projectTemplateKey required, datatype -> string
                             : name, key, description, leadAccountId, url, assigneeType, datatype -> string
                             : avatarId, issueSecurityScheme, permissionScheme, notificationScheme, categoryId,
                              datatype -> integer
                : GET - Returns the project details for a project.
                This operation can be accessed anonymously.
                     :query param: expand, datatype -> string
                                 : properties, datatype -> Array<string>

               : PUT - Updates the project details for a project.
                  :query param: expand, datatype -> string
                  :body param: projectTypeKey and projectTemplateKey required, datatype -> string
                             : name, key, description, leadAccountId, url, assigneeType, datatype -> string
                             : avatarId, issueSecurityScheme, permissionScheme, notificationScheme, categoryId,
                              datatype -> integer

              : DELETE - Deletes a project.
                   :query param: enable_undo, datatype -> boolean

        """
        if uri is not None:
            return "{}/rest/api/3/project/{}/{}".format(LOGIN.base_url, id_or_key, uri)
        else:
            if query is not None:
                return "{}/rest/api/3/project/{}?{}".format(LOGIN.base_url, id_or_key, query)
            else:
                if enable_undo is not None:
                    return "{}/rest/api/3/project/{}?enableUndo={}".format(LOGIN.base_url, id_or_key, enable_undo)
                else:
                    return "{}/rest/api/3/project/{}".format(LOGIN.base_url, id_or_key)

    @classmethod
    def issues(cls, issue_key_or_id: Optional[Any] = None, query: Optional[Any] = None,
               uri: Optional[str] = None) -> str:
        """Creates issues, delete issues,  bulk create issue, transitions.

        A transition may be applied, to move the issue or subtask to a workflow step other than
        the default start step, and issue properties set.

        :request: POST - Creates an issue or, where the option to create subtasks is enabled in Jira, a subtask.
        :param: uri, datatype -> string
        :param: query, datatype -> string
        :param: issue_key_or_id -> string or integer
                available options [bulk, createmeta]
                e.g. endpoint: /rest/api/3/issue/bulk
                e.g. endpoint /rest/api/3/issue/createmeta

                   :query param: updateHistory, datatype -> boolean
                   :body param: transition, fields, update, historyMetadata, datatype -> object
                              : properties, datatype -> Array<EntityProperty>
                              : Additional Properties, datatype -> Any

                  POST - Bulk create issue
                      Creates issues and, where the option to create subtasks is enabled in Jira, subtasks.
                      :body param: issueUpdates, datatype -> Array<IssueUpdateDetails>
                                 : Additional Properties, datatype -> Any

                  GET - Create issue metadata
                  Returns details of projects, issue types within projects, and, when requested,
                  the create screen fields for each issue type for the user.
                  :query param: projectIds, projectKeys, issuetypeIds, issuetypeNames, datatype -> Array<string>
                              : expand, datatype -> string

                  GET - Get issue
                  Return the details of an issue
                  endpoint  /rest/api/3/issue/{issueIdOrKey}
                  :query param: issue_key_or_id required
                              : fields, properties, datatype -> Array<string>
                              : fieldsByKeys, updateHistory,  datatype -> boolean
                              : expand, datatype -> string

                  PUT - Edits an issue. A transition may be applied and issue properties updated as part of the edit.
                  endpoint  /rest/api/3/issue/{issueIdOrKey}
                  :query param: issue_key_or_id required
                              : notifyUsers, overrideScreenSecurity, overrideEditableFlag, datatype -> boolean
                  :body param: transition, fields, update, historyMetadata, properties, Additional Properties,
                              datatype -> object

                  DELETE - Deletes an issue.
                  endpoint  /rest/api/3/issue/{issueIdOrKey}
                  :query param: issue_key_or_id required
                              : deleteSubtasks, datatype -> string, values = (true | false)
        """
        if uri is not None and query is None:
            return "{}/rest/api/3/issue/{}".format(LOGIN.base_url, uri)
        elif uri is not None and query is not None:
            return "{}/rest/api/3/issue/{}?{}".format(LOGIN.base_url, uri, query)
        else:
            if issue_key_or_id is not None and query is None:
                return "{}/rest/api/3/issue/{}".format(LOGIN.base_url, issue_key_or_id)
            elif issue_key_or_id is not None and query is not None:
                return "{}/rest/api/3/issue/{}?{}".format(LOGIN.base_url, issue_key_or_id, query)
            else:
                return "{}/rest/api/3/issue".format(LOGIN.base_url)


def echo(obj):
    """A Call to the Echo Class."""
    e = Echo()
    return e.echo(raw=obj)


endpoint = EndPoints()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
JiraOne is an Atlassian REST API Interface used to Generate report on certain aspect of Jira.

You can dynamically generate a report of
1. Total no of Fields used on per project or the instance
2. Total Number of Users (active, inactive as well as user type)
3. Report of Users able to Browse a certain Project, No of Issues of those Project
4. Total no of Screens used on per projects or the instance
5. Total no of workflows used on per projects or the instance
6. Total no of notification schemes on the instances and their names
7. Total no of permissions schemes on the instance and their names
8. Total no of components used on the instance and their names
9. Total no of projects on the instance
   a) Total no of project roles used per projects or the instance
   b) Total no of issue types on the projects
10. Total no of groups on the instance with their names
11. Total no of dashboards on the instance, who it is shared with
12. Total no of attachments on the project and on the instance
and many more depending on what you can come up with.
"""
from jiraone.access import LOGIN, endpoint, echo
from jiraone.jira_logs import add_log, WORK_PATH
from jiraone.reporting import PROJECT, USER, csv_writer, csv_reader, path_builder

__all__ = ["LOGIN",
           "endpoint",
           "echo",
           "add_log",
           "WORK_PATH",
           "PROJECT",
           "USER",
           "csv_writer",
           "csv_reader",
           "path_builder"]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Example methods of Generating reports

Provided herein are Report Generator Classes and Methods,
Easily generate report for the various endpoints
"""
from typing import Any, Optional, List, Iterable, Tuple, Union
from collections import deque
from jiraone import LOGIN, endpoint, add_log, WORK_PATH
import json
import csv
import sys
import os

CsvData = List[List[str]]


class Projects:
    """Get report on a Project based on user or user's attributes or groups."""

    @staticmethod
    def projects_accessible_by_users(*args: Any, project_folder: str = "Project",
                                     project_file_name: str = "project_file.csv",
                                     user_extraction_file: str = "project_extract.csv",
                                     permission: str = "BROWSE", **kwargs):
        """
        Send an argument as String equal to a value, example: status=live.

        Multiple arguments separate by comma as the first argument in the function, all other arguments should
        be keyword args that follows. This API helps to generate full user accessibility to Projects on Jira.
        It checks the users access and commits the finding to a report file. You can tweak the permission argument
        with the options mention here
        https://developer.atlassian.com/cloud/jira/platform/rest/v3/api-group-user-search
        for endpoint /rest/api/3/user/permission/search
        """
        count_start_at = 0
        headers = ["Project Key", "Project Name", "Issue Count", "Last Issue Update"]
        csv_writer(folder=project_folder, file_name=project_file_name, data=headers)

        def project():
            csv_writer(project_folder, project_file_name, data=raw)
            read_users = csv_reader(folder=project_folder, file_name=file_name)
            for user in read_users:
                account_id = user[0]
                display_name = user[2]
                active_status = user[3]
                project_key = keys
                find = LOGIN.get(endpoint.find_users_with_permission(account_id, project_key, permission))
                data = json.loads(find.content)
                if str(data) == "[]":
                    raw_vision = [display_name, f"Has {permission} Permission: False",
                                  f"Project: {name}", f"User Status: {active_status}"]
                    csv_writer(project_folder, project_file_name, data=raw_vision)
                else:
                    for d in data:
                        d_name = d["displayName"]
                        active = d["active"]
                        raw_vision = [d_name, f"Has {permission} Permission: {active}",
                                      f"Project: {name}", f"User Status: {active_status}"]

                        csv_writer(project_folder, project_file_name, data=raw_vision)

        file_name = user_extraction_file
        USER.get_all_users(file=file_name, folder=project_folder, **kwargs)
        print("Project User List Extracted")
        add_log("Project User List Extracted", "info")
        while True:
            load = LOGIN.get(endpoint.get_projects(*args, start_at=count_start_at))
            count_start_at += 50
            if load.status_code == 200:
                results = json.loads(load.content)
                for key in results["values"]:
                    keys = key["key"]
                    name = key["name"]
                    if "insight" in key:
                        insight = key["insight"]
                        if "totalIssueCount" and "lastIssueUpdateTime" in insight:
                            raw = [keys, name, f"{insight['totalIssueCount']}",
                                   f"{insight['lastIssueUpdateTime']}"]
                            project()
                        elif "totalIssueCount" in insight and "lastIssueUpdateTime" not in insight:
                            raw = [keys, name, f"{insight['totalIssueCount']}",
                                   "No data available"]
                            project()
                    else:
                        raw = [keys, name, "No data available", "No data available"]
                        project()

                if count_start_at > results["total"]:
                    print("Project Reporting Completed")
                    print("File extraction completed. Your file is located at {}"
                          .format(path_builder(path=project_folder, file_name=project_file_name)))
                    add_log("Project Reporting Completed", "info")
                    break
            else:
                sys.stderr.write("Unable to fetch data status {} ".format(load.status_code))
                add_log(f"Data retrieval failure due to {load.reason}", "error")
                sys.exit(1)

    @staticmethod
    def dashboards_shared_with(dashboard_folder: str = "Dashboard",
                               dashboard_file_name: str = "dashboard_file.csv",
                               **kwargs):
        """
        Retrieve the Dashboard Id/Name/owner and who it is shared with.

        The only requirement is that the user querying this API should have access to all
        the Dashboard available else it will only return dashboard where the user's view access is allowed.
        """
        count_start_at = 0
        dash_list = deque()  # get the id of dashboard search
        put_list = deque()  # arrange the different values associated with dashboard
        dump_list = deque()  # dump the results into our csv writer
        headers = ["Dashboard Id", "Dashboard Name", "Owner", "Shared Permission"]
        csv_writer(folder=dashboard_folder, file_name=dashboard_file_name, data=headers, **kwargs)

        while True:
            load = LOGIN.get(endpoint.search_for_dashboard(start_at=count_start_at))
            count_start_at += 50
            if load.status_code == 200:
                results = json.loads(load.content)
                for value in results["values"]:
                    valid = value["id"]
                    name = value["name"]
                    raw = [valid, name]
                    dash_list.append(raw)

                print("Dashboard List Extracted")
                add_log("Dashboard List Extracted", "info")

                def dash_run():
                    for pem in shared_permission:
                        if "role" in pem:
                            role = f"Role: {pem['role']['name']}"
                            put_list.append(role)
                        if "group" in pem:
                            group = f"Group: {pem['group']['name']}"
                            put_list.append(group)
                        if "type" in pem:
                            types = pem["type"]
                            if types == "project":
                                project = f"Project Name: {pem['project']['name']}"
                                put_list.append(project)
                            if types == "loggedin" or types == "global":
                                put_list.append(types)

                    extract = [d for d in put_list]
                    raw = [value[0], names, owner, extract]
                    dump_list.append(raw)

                for value in dash_list:
                    init = LOGIN.get(endpoint.get_dashboard(value[0]))
                    if init.status_code == 200:
                        expanse = json.loads(init.content)
                        names = expanse["name"]
                        if "owner" in expanse:
                            owner = expanse["owner"]["displayName"]
                        shared_permission = expanse["sharePermissions"]
                        dash_run()
                        put_list = deque()

                csv_writer(dashboard_folder, dashboard_file_name, data=dump_list, mark="many")

                if count_start_at > results["total"]:
                    print("Dashboard Reporting Completed")
                    print("File extraction completed. Your file is located at {}"
                          .format(path_builder(path=dashboard_folder, file_name=dashboard_file_name)))
                    add_log("Dashboard Reporting Completed", "info")
                    break

    @staticmethod
    def get_all_roles_for_projects(roles_folder: str = "Roles",
                                   roles_file_name: str = "roles_file.csv",
                                   user_extraction: str = "role_users.csv",
                                   **kwargs):
        """Get the roles available in a project and which user is assigned to which
        role within the project"""
        count_start_at = 0
        headers = ["Project Id ", "Project Key", "Project Name", "Project roles", "User AccountId", "User DisplayName",
                   "User role in Project"]
        csv_writer(folder=roles_folder, file_name=roles_file_name, data=headers, **kwargs)

        # get extraction of projects data
        def role_on() -> Any:
            project_role_list = deque()
            for keys in results["values"]:
                role_list = deque()
                key = keys["key"]
                pid = keys["id"]
                name = keys["name"]
                roles = LOGIN.get(endpoint.get_roles_for_project(pid))
                print("Extracting Project Keys {}".format(key))
                add_log("Extracting Project Keys {}".format(key), "info")

                # find out what roles exist within each project
                def role_over() -> Tuple:
                    if roles.status_code == 200:
                        extract = json.loads(roles.content)
                        only_keys = extract.keys()
                        casting = list(only_keys)
                        only_values = extract.items()
                        puller = list(only_values)
                        z = [d for d in puller]
                        return casting, z

                caster = role_over()
                raw = [pid, key, name, caster[0]]
                role_list.append(raw)
                add_log("Appending data to List Queue", "info")

                for user in read_users:
                    account_id = user[0]
                    display_name = user[2]

                    # extract the user role using the appropriate accountId
                    def get_user_role(user_data) -> Any:
                        for users in user_data[1]:
                            check = LOGIN.get(users[1])
                            if check.status_code == 200:
                                result_data = json.loads(check.content)
                                actors = result_data["actors"]
                                for act in actors:
                                    if "actorUser" in act:
                                        if account_id == act["actorUser"]["accountId"]:
                                            res = f"Role Name: {result_data['name']}"
                                            project_role_list.append(res)

                        # function to write collected data into a file
                        def pull_data() -> Any:
                            role_puller = [j for j in project_role_list]
                            project_id = data[0]
                            project_key = data[1]
                            project_name = data[2]
                            project_roles = data[3]
                            raw_dump = [project_id, project_key, project_name, project_roles, account_id, display_name,
                                        role_puller]
                            csv_writer(folder=roles_folder, file_name=roles_file_name, data=raw_dump)

                        for data in role_list:
                            pull_data()

                    get_user_role(caster)
                    project_role_list = deque()

        USER.get_all_users(folder=roles_folder, file=user_extraction, **kwargs)
        read_users = csv_reader(folder=roles_folder, file_name=user_extraction, **kwargs)
        while True:
            init = LOGIN.get(endpoint.get_projects(start_at=count_start_at))
            if init.status_code == 200:
                print("Project Extraction")
                results = json.loads(init.content)
                add_log("Project Extraction Initiated", "info")
                if count_start_at > results["total"]:
                    break
                if results["total"] > 0:
                    role_on()
            count_start_at += 50

        print("File extraction completed. Your file is located at {}".format(path_builder
                                                                             (path=roles_folder,
                                                                              file_name=roles_file_name)))
        add_log("File extraction completed", "info")

    def get_attachments_on_projects(self, attachment_folder: str = "Attachment",
                                    attachment_file_name: str = "attachment_file.csv",
                                    **kwargs):
        """Return all attachments on a Project or Projects

        Get the size of attachments on an Issues, count those attachments collectively and return the total number
        on all Projects searched. JQL is used as a means to search for the project.
        """
        import re
        attach_list = deque()
        count_start_at = 0
        headers = ["Project Id", "Project Key", "Project Name", "Issue Key", "Attachment size",
                   "Attachment Type", "Created on by user", "Attachment url"]
        csv_writer(folder=attachment_folder, file_name=attachment_file_name, data=headers, **kwargs)

        def pull_attachment_sequence() -> Optional:
            nonlocal attach_list
            for issues in result_data["issues"]:
                keys = issues["key"]
                get_issue_keys = LOGIN.get(endpoint.issues(issue_key_or_id=keys))
                if get_issue_keys.status_code == 200:
                    key_data = json.loads(get_issue_keys.content)
                    data = key_data["fields"]
                    if "project" or "attachment" in data:
                        project_id = data["project"]["id"]
                        project_name = data["project"]["name"]
                        project_key = data["project"]["key"]
                        attachment = data["attachment"]
                        if attachment is not None:
                            for attach in attachment:
                                if "author" in attach:
                                    display_name = attach["author"]["displayName"]

                                def pull_attachment() -> Optional:
                                    created = attach["created"]  # datetime need to convert it
                                    attachment_size = attach["size"]  # in bytes, need to convert to mb
                                    mime_type = attach["mimeType"]
                                    attachment_url = attach["content"]

                                    calc_size = self.byte_converter(attachment_size)

                                    def date_converter(val) -> str:
                                        """split the datetime value and output a string."""
                                        get_date_time = val.split("T")
                                        get_am_pm = get_date_time[1].split(".")
                                        return f"Created on {get_date_time[0]} {get_am_pm[0]}"

                                    calc_date = date_converter(created)
                                    pull = [project_id, project_key, project_name, keys, calc_size, mime_type,
                                            f"{calc_date} by {display_name}", attachment_url]
                                    attach_list.append(pull)

                                pull_attachment()

            raw_data = [x for x in attach_list]
            csv_writer(attachment_folder, attachment_file_name, data=raw_data, mark="many")
            attach_list.clear()

        while True:
            get_issue = LOGIN.get(endpoint.search_issues_jql(start_at=count_start_at, **kwargs))
            if get_issue.status_code == 200:
                result_data = json.loads(get_issue.content)
                if count_start_at > result_data["total"]:
                    print("Attachment extraction completed")
                    add_log("Attachment extraction completed", "info")
                    break

                print("Attachment extraction processing")
                add_log("Attachment extraction processing", "info")
                pull_attachment_sequence()

            count_start_at += 50

        def grade_and_sort() -> Union[float, int]:
            for node in read_file:
                pattern = re.compile(r"(\d*?[0-9]\.[0-9]*)", re.I)
                if pattern is not None:
                    if pattern.search(node[4]):
                        attach_list.append(pattern.search(node[4]).group())

            retain = [float(s) for s in attach_list]
            calc_sum = sum(retain)
            return calc_sum

        def re_write():
            calc_made = grade_and_sort()
            attach_list.clear()
            rd_file = read_file
            # sort the file using the issue_key column
            sorts = sorted(rd_file, key=lambda row: row[3], reverse=False)
            for i in sorts:
                _project_id = i[0]
                _project_key = i[1]
                _project_name = i[2]
                _issue_key = i[3]
                _attach_size = i[4]
                _attach_type = i[5]
                _created_by = i[6]
                _attach_url = i[7]
                raw_data_file = [_project_id, _project_key, _project_name, _issue_key,
                                 _attach_size, _attach_type, _created_by, _attach_url]
                csv_writer(attachment_folder, attachment_file_name, data=raw_data_file)

            # lastly we want to append the total sum of attachment size.
            raw_data_file = ["", "", "", "", "Total Size: {:.2f} MB".format(calc_made), "", "", ""]
            csv_writer(attachment_folder, attachment_file_name, data=raw_data_file)

        read_file = csv_reader(attachment_folder, attachment_file_name, skip=True)
        csv_writer(attachment_folder, attachment_file_name, data=headers, mode="w")
        re_write()

        print("File extraction completed. Your file is located at {}".format(path_builder
                                                                             (path=attachment_folder,
                                                                              file_name=attachment_file_name)))
        add_log("File extraction completed", "info")

    @staticmethod
    def byte_converter(val) -> str:
        """1 Byte = 8 Bits.

        using megabyte MB, value is 1000^2
        mebibyte MiB, value is 1024^2
        Therefore val = val / MB
        """
        byte_size = val
        mega_byte = 1000 * 1000
        visor = byte_size / mega_byte
        return "Size: {:.2f} MB".format(visor)


class Users:
    """
    Below methods helps to Generate the No of Users on Jira Cloud

    You can customize it to determine which user you're looking for.
    >> The below method here displays active or inactive users, so you'll be getting all users
    :param: pull (options)
            -> both: pulls out inactive and active users
            -> active: pulls out only active users
            -> inactive: pulls out inactive users
    :param: user_type (options)
            -> atlassian: a normal Jira Cloud user
            -> customer: this will be your JSM customers
            -> app: this will be the bot users for any Cloud App
            -> unknown: as the name suggest unknown user type probably from oAuth
    """
    user_list = deque()

    def get_all_users(self, pull: str = "both", user_type: str = "atlassian",
                      file: str = None, folder: str = Any, **kwargs) -> Any:
        """Generates a list of users."""
        count_start_at = 0
        validate = LOGIN.get(endpoint.myself())

        while validate.status_code == 200:
            extract = LOGIN.get(endpoint.search_users(count_start_at))
            results = json.loads(extract.content)
            self.user_activity(pull, user_type, results)
            count_start_at += 50
            print("Current Record - At Row", count_start_at)
            add_log(f"Current Record - At Row {count_start_at}", "info")

            if str(results) == "[]":
                break
        else:
            sys.stderr.write("Unable to connect to {} - Login Failed...".format(LOGIN.base_url))
            add_log(f"Login Failure on {LOGIN.base_url}, due to {validate.reason}", "error")
            sys.exit(1)

        if file is not None:
            self.report(category=folder, filename=file)

    def report(self, category: str = Any, filename: str = "users_report.csv") -> Optional:
        """Creates a user report file in CSV format."""
        read = [d for d in self.user_list]
        csv_writer(folder=category, file_name=filename, data=read, mark="many")
        add_log(f"Generating report file on {filename}", "info")

    def user_activity(self, status: str = Any, account_type: str = Any, results: List = Any) -> Any:
        """Determines users activity."""

        # get both active and inactive users
        def stack(c: Any, f: Any, s: Any) -> Any:
            if status == "both":
                if s["accountType"] == account_type:
                    return c.user_list.append(f)
            elif status == "active":
                if s["accountType"] == account_type and s["active"] is True:
                    return c.user_list.append(f)
            elif status == "inactive":
                if s["accountType"] == account_type and s["active"] is False:
                    return c.user_list.append(f)

        for each_user in results:
            list_user = [each_user["accountId"], each_user["accountType"], each_user["displayName"],
                         each_user["active"]]
            stack(self, list_user, each_user)

    def get_all_users_group(self, group_folder: str = "Groups", group_file_name: str = "group_file.csv",
                            user_extraction_file: str = "group_extraction.csv", **kwargs):
        """Get all users and the groups associated to them on the Instance."""
        headers = ["Name", "AccountId", "Groups", "User status"]
        csv_writer(folder=group_folder, file_name=group_file_name, data=headers, **kwargs)
        file_name = user_extraction_file
        self.get_all_users(file=file_name, folder=group_folder, **kwargs)
        reader = csv_reader(file_name=file_name, folder=group_folder, **kwargs)
        for user in reader:
            account_id = user[0]
            display_name = user[2]
            active_status = user[3]
            load = LOGIN.get(endpoint.get_user_group(account_id))
            results = json.loads(load.content)
            get_all = [d["name"] for d in results]
            raw = [display_name, account_id, get_all, active_status]
            csv_writer(folder=group_folder, file_name=group_file_name, data=raw)

        print("File extraction completed. Your file is located at {}"
              .format(path_builder(path=group_folder, file_name=group_file_name)))
        add_log("Get Users group Completed", "info")


def path_builder(path: str = "Report", file_name: str = Any, **kwargs):
    """Builds a dir path and file path in a directory."""
    base_dir = os.path.join(WORK_PATH, path)
    if not os.path.exists(base_dir):
        os.mkdir(base_dir)
        add_log(f"Building Path {path}", "info")
    base_file = os.path.join(base_dir, file_name)
    return base_file


def csv_writer(folder: str = Any, file_name: str = Any, data: Iterable = object,
               mark: str = "single", mode: str = "a+", **kwargs) -> Any:
    """Reads and writes to a file, single or multiple rows."""
    file = path_builder(path=folder, file_name=file_name)
    if mode:
        with open(file, mode) as f:
            write = csv.writer(f, delimiter=",")
            if mark == "single":
                write.writerow(data)
            if mark == "many":
                write.writerows(data)
            add_log(f"Writing to file {file_name}", "info")


def csv_reader(folder: str = Any, file_name: str = Any, mode: str = "r",
               skip: bool = False, **kwargs) -> CsvData:
    """Reads a CSV file and returns a list comprehension of the data."""
    file = path_builder(path=folder, file_name=file_name)
    if mode:
        with open(file, mode) as f:
            read = csv.reader(f, delimiter=",")
            if skip is True:
                next(read, None)
            load = [d for d in read]
            add_log(f"Read file {file_name}", "info")
            return load


USER = Users()
PROJECT = Projects()